!
! Module: attbin_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task bins up spacecraft attitude.
! 
! Comments: 
!	TODO:
!	- write a proper test harness for time-sorting.
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module attbin_mod

  use oal
  use dal
  use param  
  use errorHandling
  use caloaldefs	! for SpacecraftAttitudeType

  use binned_att

  use array_utils
  use constants
  use dal_aux
  use ssc_misc
  use sort_mod		! for quickSort

  implicit none 

  private 

  public attbin

  ! Command-line parameters:
  character(256) ::&
    atthkSetName='',&
    eventListSetName='',&
    binnedAttSetName='',&
    pntKeywordSetName=''
  character(20) ::&
    startDateTimeString='',&
    stopDateTimeString='',&
    sourceOfAttInfo='',&
    sourceOfOdfAttInfo='',&
    sourceOfAttHkAttInfo='',&
    sourceOfTimeInfo='',&
    userTimeInfoStyle=''
  real(double) ::&
    samplePeriod=0.0d0,&
    maxDeltaAngleArcsec=0.0d0,&
    userStartTimeStamp=0.0d0,&
    userStopTimeStamp=0.0d0
  logical ::&
    useMedianForPnt=.false.,&
    calcMeanInBins=.false.,&
    inputAsTimeStamp=.false.

contains 

  !---------------------------------------------------------------------
  subroutine attbin
  !---------------------------------------------------------------------
  ! This task reads an event list to get a sequence of times; the s/c attitude is obtained for each new time; if the attitude departs from the last-sampled attitude by more than a given amount, a new sample is made. The samples are stored at the end of the processing.

    implicit none 

    real(single), parameter :: nominalFovRadiusArcmin = 15.0&
         , timeSecIntervalNegligible = 0.0001

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: raCol, decCol, aposCol
    real(double) ::&
      maxDeltaAngle, startTimeStamp, stopTimeStamp,&
! startTimeFirstBin,&
      meanVal, medianVal, denom
    integer :: numTimeSamples, i, numBins, rowOfFirstGoodAtt, binNum, iHi, iLo
    real(double), pointer :: times(:)
    ! real(double), allocatable :: times_sort(:)
    type(attBinType), allocatable :: binnedAtt(:)
    type(SpacecraftAttitudeType) :: scAttitude, maxDelta, pntDeg
    type(SpacecraftAttitudeType), allocatable :: scAttitudeVec(:)
    logical :: attIsGood, attIsNew
    logical, allocatable :: attIsGoodVec(:)
    integer, allocatable :: firstIInBin(:)
    character(7) :: pntStyle
    real(double), allocatable :: angleDeg(:)

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    maxDeltaAngle = maxDeltaAngleArcsec / 3600.0 * PiOn180
    maxDelta%dec  = maxDeltaAngle
    maxDelta%apos = maxDeltaAngle / (nominalFovRadiusArcmin * PiOn180 / 60.0)

    select case(sourceOfAttInfo)
    case('odf')
      select case(sourceOfOdfAttInfo)
      case('ahf')
        call OAL_setState(attitudeFromAhf = .true.)       
      case('om')
        call OAL_setState(attitudeFromAhf = .false.)       
      case default
call error('badOdfAttitudeSource',"")
      end select

      select case(sourceOfTimeInfo)
      case('eventlist')
        ! In this case must get mean/median attitude from somewhere else,
        ! because it is too fiddly to calculate them (especially the median)
        ! from an uneven sequence of time samples.

        set = dataSet(pntKeywordSetName, READ)
        if (useMedianForPnt) then
          select case(sourceOfOdfAttInfo)
          case('ahf')
            pntDeg%ra   = real64Attribute(set,"MAHFRA")
            pntDeg%dec  = real64Attribute(set,"MAHFDEC")
            pntDeg%apos = real64Attribute(set,"MAHFPA")
          case('om')
            pntDeg%ra   = real64Attribute(set,"MOMRA")
            pntDeg%dec  = real64Attribute(set,"MOMDEC")
            pntDeg%apos = real64Attribute(set,"MOMPA")
          case default
call error('badOdfAttitudeSource',"")
          end select

          pntStyle = 'MEDIAN'

        else ! use the mean for pnt
          select case(sourceOfOdfAttInfo)
          case('ahf')
            pntDeg%ra   = real64Attribute(set,"AAHFRA")
            pntDeg%dec  = real64Attribute(set,"AAHFDEC")
            pntDeg%apos = real64Attribute(set,"AAHFPA")
          case('om')
            pntDeg%ra   = real64Attribute(set,"AOMRA")
            pntDeg%dec  = real64Attribute(set,"AOMDEC")
            pntDeg%apos = real64Attribute(set,"AOMPA")
          case default
call error('badOdfAttitudeSource',"")
          end select

          pntStyle = 'MEAN'
        end if
        call release(set)

        maxDelta%ra   = maxDeltaAngle / cos(pntDeg%dec * PiOn180)

        set = dataSet(eventListSetName, READ)
        tab = table(set, 'EVENTS')
        numTimeSamples = numberOfRows(tab)
if (numTimeSamples <= 0) call error('noRowsInEventList',"")
        times => real64Data(column(tab, 'TIME'))

        ! Sort the time (modify times()).  (eg. SPR-3651)
        ! Note: In general, the time order in an event list is not
        !  strictly in the ascending order.  Therefore the following
        !  procedure with quickSort() is essential.
        !  It seems this is also achieved in the separate task, etimedither.
        !allocate(times_sort(0:numTimeSamples-1))
        !times_sort = times
        call message(appMsg, Verbose, 'QuickSort('//trim(eventListSetName)//') started.')
        call quickSort(times)		! times is modified.
        ! in ssclib/sort_mod
        call message(appMsg, Verbose, 'QuickSort ended.')
        if (any(times(0:numTimeSamples-2) > times(1:numTimeSamples-1)))&
call error('badEventTimeOrder',"Contact the code developer!")
        ! The above check "if (any(...))" should be unnecessary, providing
        ! quickSort works properly.

        allocate(binnedAtt(numTimeSamples))
!        binnedAtt%fracexp = 1.0
call initializeBinnedAtt(binnedAtt)
        ! Now, 
        !  binnedAtt(1:numTimeSamples)%isGood() = (F, ..., F)
        !  binnedAtt%_everythingelse_() = (0.0d0, ..., 0.0d0)
        ! Below, this will change into, for example, 
        !  binnedAtt%isGood() = (T, T, T, T, ...)
        !  binnedAtt%_else_() = (6, 2, 1, 8, ...)
        ! where the original data (NOT array, actualy) would be, for example, 
        !  orig%isGood[]      = [F, F, T, T, T, T, T, T, T, T, T, T, T, ...]
        !  originalatt%[]     =[0, 0, 6,6.1,5.9,6,2,1.9,2.1,1, 8,8.1,8.1,...]
        !  attIsNew[]         = [F, F, T, F, F, F, T, F, F, T, T, F, F, ...]
        ! In that sense, binnedAtt is a kind of "degenerated" array, meaning
        ! many bins at the tail will remain as the initial ones (F/0.0),
        ! which will be stripped when  call writeBinnedAttitude(...) (See below).

        ! Note that the variable numBins below is the index for
        ! the "significant" bins in binnedAtt .

        ! Seek and set the first good bin.
        rowOfFirstGoodAtt = numTimeSamples
        do i = 0, numTimeSamples-1
          call binUpAttitudeFromOdf(times(i), maxDelta, scAttitude&
            , attIsGood, attIsNew)

          if (attIsNew) then
            numBins = 1
!            startTimeFirstBin = times(i)
            binnedAtt(numBins)%att = scAttitude
            binnedAtt(numBins)%isGood = attIsGood	! Must be always .true.
          end if

          if (attIsGood) then
            rowOfFirstGoodAtt = i
        exit
          end if
        end do
if (rowOfFirstGoodAtt >= numTimeSamples) call error('noGoodAttitudes',"")

        ! Set the following bins.
        do i = rowOfFirstGoodAtt + 1, numTimeSamples-1
          errStr = sprintEvery(i, numTimeSamples, fraction=0.1)
          ! in ssclib/ssc_misc.

          if (errStr .ne. '') call message(appMsg, Verbose)

          if (times(i) - times(i-1) >= 0.0001) then 
            call binUpAttitudeFromOdf(times(i), maxDelta, scAttitude&
              , attIsGood, attIsNew)

            if (attIsNew) then
              numBins = numBins + 1
!              binnedAtt(numBins-1)%binEndTime = times(i)
              binnedAtt(numBins)%binStartTime = times(i)
              binnedAtt(numBins)%att = scAttitude
              binnedAtt(numBins)%isGood = attIsGood
            end if
          end if
        end do
!        binnedAtt(numBins)%binEndTime = nearest(times(numTimeSamples-1),1.0)

        call release(set)

      case('user') ! sourceOfTimeInfo=user
        if (inputAsTimeStamp) then
          startTimeStamp = userStartTimeStamp
          stopTimeStamp  = userStopTimeStamp
        else
          startTimeStamp = toTimeTag(startDateTimeString)
          stopTimeStamp  = toTimeTag(stopDateTimeString)
        end if

if (stopTimeStamp <= startTimeStamp) call error('badStartStopTimes',"")
        numTimeSamples = floor((stopTimeStamp - startTimeStamp) / samplePeriod)
!if (numTimeSamples <= 0) call error('noTimeSamples',"")
if (numTimeSamples <= 0) numTimeSamples = 1
        allocate(times(0:numTimeSamples-1))
        allocate(scAttitudeVec(0:numTimeSamples-1))
        allocate(attIsGoodVec(0:numTimeSamples-1))
        times = (/(startTimeStamp + i * samplePeriod, i=0,numTimeSamples-1)/)

        do i = 0, numTimeSamples-1
          call getEventAttitude(times(i), scAttitudeVec(i), attIsGoodVec(i))
        end do

        call calcMeanMedian(scAttitudeVec%ra, attIsGoodVec, meanVal, medianVal)
        if (useMedianForPnt) then
          pntDeg%ra = medianVal / PiOn180
        else
          pntDeg%ra = meanVal / PiOn180
        end if

        call calcMeanMedian(scAttitudeVec%dec, attIsGoodVec, meanVal, medianVal)
        if (useMedianForPnt) then
          pntDeg%dec = medianVal / PiOn180
        else
          pntDeg%dec = meanVal / PiOn180
        end if

        call calcMeanMedian(scAttitudeVec%apos, attIsGoodVec, meanVal,medianVal)
        if (useMedianForPnt) then
          pntDeg%apos = medianVal / PiOn180
        else
          pntDeg%apos = meanVal / PiOn180
        end if

        if (useMedianForPnt) then
          pntStyle = 'MEDIAN'
        else
          pntStyle = 'MEAN'
        end if

        maxDelta%ra   = maxDeltaAngle / cos(pntDeg%dec * PiOn180)

        allocate(binnedAtt(numTimeSamples))
!        binnedAtt%fracexp = 1.0
call initializeBinnedAtt(binnedAtt)
        ! NOTE: See the comments below 1st  call initializeBinnedAtt(binnedAtt)

        rowOfFirstGoodAtt = numTimeSamples
        do i = 0, numTimeSamples-1
          call checkIfNewAttBin(scAttitudeVec(i), maxDelta&
            , attIsGoodVec(i), attIsNew)

          if (attIsNew) then
            numBins = 1
!            startTimeFirstBin = times(i)
            binnedAtt(numBins)%att = scAttitudeVec(i)
            binnedAtt(numBins)%isGood = attIsGoodVec(i)
          end if

          if (attIsGoodVec(i)) then
            rowOfFirstGoodAtt = i
        exit
          end if
        end do
if (rowOfFirstGoodAtt >= numTimeSamples) call error('noGoodAttitudes',"")

        do i = rowOfFirstGoodAtt + 1, numTimeSamples-1
          call checkIfNewAttBin(scAttitudeVec(i), maxDelta&
            , attIsGoodVec(i), attIsNew)

          if (attIsNew) then
            numBins = numBins + 1
!            binnedAtt(numBins-1)%binEndTime = times(i)
            binnedAtt(numBins)%binStartTime = times(i)
            binnedAtt(numBins)%att = scAttitudeVec(i)
            binnedAtt(numBins)%isGood = attIsGoodVec(i)
          end if
        end do
!        binnedAtt(numBins)%binEndTime = nearest(times(numTimeSamples-1),1.0)

        deallocate(times)

      case default ! sourceOfTimeInfo
call error('badTimeSource',"")
      end select ! sourceOfTimeInfo

    case('atthk') ! sourceOfAttInfo
      set = dataSet(atthkSetName, READ)
      tab = table(set, 'ATTHK')
      numTimeSamples = numberOfRows(tab)
if (numTimeSamples <= 0) call error('noRowsInAtthkFile',"")
      times => real64Data(column(tab, 'TIME'))
      allocate(scAttitudeVec(0:numTimeSamples-1))

      select case(sourceOfAttHkAttInfo)
      case('ahf')
        if (useMedianForPnt) then
          pntDeg%ra   = real64Attribute(set,"MAHFRA")
          pntDeg%dec  = real64Attribute(set,"MAHFDEC")
          pntDeg%apos = real64Attribute(set,"MAHFPA")
          pntStyle = 'MEDIAN'
        else ! use the mean for pnt
          pntDeg%ra   = real64Attribute(set,"AAHFRA")
          pntDeg%dec  = real64Attribute(set,"AAHFDEC")
          pntDeg%apos = real64Attribute(set,"AAHFPA")
          pntStyle = 'MEAN'
        end if

        raCol   = column(tab, 'AHFRA')
        decCol  = column(tab, 'AHFDEC')
        aposCol = column(tab, 'AHFPA')

      case('om')
       if (useMedianForPnt) then
          pntDeg%ra   = real64Attribute(set,"MOMRA")
          pntDeg%dec  = real64Attribute(set,"MOMDEC")
          pntDeg%apos = real64Attribute(set,"MOMPA")
          pntStyle = 'MEDIAN'
        else ! use the mean for pnt
          pntDeg%ra   = real64Attribute(set,"AOMRA")
          pntDeg%dec  = real64Attribute(set,"AOMDEC")
          pntDeg%apos = real64Attribute(set,"AOMPA")
          pntStyle = 'MEAN'
        end if

        raCol   = column(tab, 'OMRA')
        decCol  = column(tab, 'OMDEC')
        aposCol = column(tab, 'OMPA')
      case default ! sourceOfAttHkAttInfo
call error('badAtthkAttitudeSource',"")
      end select

      maxDelta%ra = maxDeltaAngle / cos(pntDeg%dec * PiOn180)

      allocate(angleDeg(numTimeSamples))
      allocate(firstIInBin(numTimeSamples))
      call readColDataToFixed(raCol,   angleDeg)   ! in
      scAttitudeVec%ra   = angleDeg * PiOn180
      call readColDataToFixed(decCol,  angleDeg)  ! ssclib/
      scAttitudeVec%dec  = angleDeg * PiOn180
      call readColDataToFixed(aposCol, angleDeg) ! dal_aux.
      scAttitudeVec%apos = angleDeg * PiOn180

      allocate(binnedAtt(numTimeSamples))
!      binnedAtt%fracexp = 1.0
call initializeBinnedAtt(binnedAtt)
      ! NOTE: See the comments below 1st  call initializeBinnedAtt(binnedAtt)

      rowOfFirstGoodAtt = numTimeSamples
      do i = 0, numTimeSamples-1
        if (isNull(raCol, i).or.isNull(decCol, i).or.isNull(aposCol, i)) then
          attIsGood = .false.
        else
          attIsGood = .true.
        end if

!write (*, '(i5.5)') i
        call checkIfNewAttBin(scAttitudeVec(i), maxDelta&
          , attIsGood, attIsNew)

        if (attIsNew) then
          numBins = 1
!print *, '*********************************'
!print *, numBins
!          startTimeFirstBin = times(i)
          binnedAtt(numBins)%att = scAttitudeVec(i)
          binnedAtt(numBins)%isGood = attIsGood
          firstIInBin(numBins) = i
        end if
!write (*, '(2f10.4)') (scAttitudeVec(i)%ra-scAttitudeVec(1)%ra)*3600.0/PiOn180&
!  , (binnedAtt(numBins)%att%ra-scAttitudeVec(1)%ra)*3600.0/PiOn180

        if (attIsGood) then
          rowOfFirstGoodAtt = i
      exit
        end if
      end do
if (rowOfFirstGoodAtt >= numTimeSamples) call error('noGoodAttitudes',"")

      do i = rowOfFirstGoodAtt + 1, numTimeSamples-1
        if (isNull(raCol, i).or.isNull(decCol, i).or.isNull(aposCol, i)) then
          attIsGood = .false.
        else
          attIsGood = .true.
        end if

!write (*, '(i5.5)') i
        call checkIfNewAttBin(scAttitudeVec(i), maxDelta&
          , attIsGood, attIsNew)

        if (attIsNew) then
          numBins = numBins + 1
!print *, '*********************************'
!print *, numBins
!          binnedAtt(numBins-1)%binEndTime = times(i)
          binnedAtt(numBins)%binStartTime = times(i)
          binnedAtt(numBins)%att = scAttitudeVec(i)
          binnedAtt(numBins)%isGood = attIsGood
          firstIInBin(numBins) = i
        end if
!write (*, '(2f10.4)') (scAttitudeVec(i)%ra-scAttitudeVec(1)%ra)*3600.0/PiOn180&
!  , (binnedAtt(numBins)%att%ra-scAttitudeVec(1)%ra)*3600.0/PiOn180
      end do
!      binnedAtt(numBins)%binEndTime = nearest(times(numTimeSamples-1),1.0)

      if (calcMeanInBins) then
        do binNum = 1, numBins-1
          iLo = firstIInBin(binNum)
          iHi = firstIInBin(binNum+1)-1
        if (.not. binnedAtt(binNum)%isGood) cycle

          denom = real(1 + iHi - iLo)
!print *, 'iLo, iHi, denom', iLo, iHi, denom
if (denom <= 0.0) call error('bug0',"Please contact the code developer!")
          binnedAtt(binNum)%att%ra   = sum(scAttitudeVec(iLo:iHi)%ra)   / denom
          binnedAtt(binNum)%att%dec  = sum(scAttitudeVec(iLo:iHi)%dec)  / denom
          binnedAtt(binNum)%att%apos = sum(scAttitudeVec(iLo:iHi)%apos) / denom
!print *, 'iLo, iHi, denom, binnedAtt(binNum)%att%ra', iLo, iHi, denom, binnedAtt(binNum)%att%ra
        end do

        binNum = numBins
        iLo = firstIInBin(binNum)
        iHi = numTimeSamples-1
        if (binnedAtt(binNum)%isGood) then
          denom = real(1 + iHi - iLo)
!print *, 'iLo, iHi, denom', iLo, iHi, denom
if (denom <= 0.0) call error('bug1',"Please contact the code developer!")
          binnedAtt(binNum)%att%ra   = sum(scAttitudeVec(iLo:iHi)%ra)   / denom
          binnedAtt(binNum)%att%dec  = sum(scAttitudeVec(iLo:iHi)%dec)  / denom
          binnedAtt(binNum)%att%apos = sum(scAttitudeVec(iLo:iHi)%apos) / denom
!print *, 'iLo, iHi, denom, binnedAtt(binNum)%att%ra', iLo, iHi, denom, binnedAtt(binNum)%att%ra
        end if

      end if

      deallocate(scAttitudeVec)
      call release(set)

!print *, ''
!print *, maxDelta%ra, maxDelta%dec, maxDelta%apos
!print *, pntDeg%dec, cos(pntDeg%dec)

    case default ! sourceOfAttInfo
call error('badAttitudeSource',"")
    end select ! sourceOfAttInfo


    ! numBins remains as the final index of the significant rows.
    write(errStr, '(i6, a)') numBins, 'rows in the output attitude timeseries.'
    call message(appMsg, Verbose)

!    call writeBinnedAttitude(binnedAttSetName, binnedAtt(1:numBins)&
!      , startTimeFirstBin, pntDeg, maxDelta, pntStyle&
!      , (/.false., .false./))
    call writeBinnedAttitude(binnedAttSetName, 'ATT_BINS', binnedAtt(1:numBins)&
      , pntDeg, maxDelta, pntStyle)
    ! in binned_att

!********** later, add a parameter to allow the user to specify the table name.

  end subroutine attbin

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    sourceOfAttInfo = stringParameter('attsource')
    select case(sourceOfAttInfo)
    case('odf')
      sourceOfOdfAttInfo = stringParameter('odfattsource')
      select case(sourceOfOdfAttInfo)
      case('ahf')
      case('om')
      case default
call error('badOdfAttitudeSource',"")
      end select

      sourceOfTimeInfo = stringParameter('timesource')
      select case(sourceOfTimeInfo)
      case('eventlist')
        eventListSetName = stringParameter('eventset')
        pntKeywordSetName = stringParameter('pntkwdset')
      case('user')
        userTimeInfoStyle = stringParameter('usertimestyle')
        if (userTimeInfoStyle == 'timestamp') then
          inputAsTimeStamp = .true.
        else if (userTimeInfoStyle == 'datetimestr') then
          inputAsTimeStamp = .false.
        else
call error('badUserTimeStyle',"")
        end if

        if (inputAsTimeStamp) then
          userStartTimeStamp = realParameter('tstart')
          userStopTimeStamp  = realParameter('tstop')
        else
          startDateTimeString = stringParameter('datestrstart')
          stopDateTimeString  = stringParameter('datestrstop')
        end if

        samplePeriod = realParameter('tstep')
      case default
call error('badTimeSource',"")
      end select

    case('atthk')
      atthkSetName = stringParameter('atthkset')

      sourceOfAttHkAttInfo = stringParameter('atthkattsource')
      select case(sourceOfAttHkAttInfo)
      case('ahf')
      case('om')
      case default
call error('badAtthkAttitudeSource',"")
      end select

!      calcMeanInBins = booleanParameter('calcmeaninbin')
calcMeanInBins = .true.

    case default
call error('badAttitudeSource',"")
    end select

    binnedAttSetName = stringParameter('binnedattset')
    maxDeltaAngleArcsec = realParameter('maxdelta')
    useMedianForPnt = booleanParameter('withmedianpnt')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine checkIfNewAttBin(att, maxDelta, attIsGood&
  , attIsNew)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in)  :: att, maxDelta
    logical,                      intent(in)  :: attIsGood
    logical,                      intent(out) :: attIsNew

    logical, save ::&
      noGoodAttsYet = .true., lastAttWasGood = .false.,&
      isFirstCall = .true.
    type(SpacecraftAttitudeType), save :: baselineAtt
    logical :: raIsNew, decIsNew, paIsNew

    !-------------------------------------------------------------------
    ! Start:

    attIsNew = .false. ! default

    if (attIsGood) then
      if (noGoodAttsYet) then
        baselineAtt = att
        attIsNew = .true.

      else ! some good atts have already been found, thus guaranteeing
      ! the existence of a baseline to test against.
        raIsNew  = .false.
        decIsNew = .false.
        paIsNew  = .false.

        if (abs(att%ra - baselineAtt%ra) > maxDelta%ra) then
          baselineAtt%ra = att%ra
          raIsNew = .true.
        endif
        if (abs(att%dec - baselineAtt%dec) > maxDelta%dec) then 
          baselineAtt%dec = att%dec
          decIsNew = .true.
        endif
        if (abs(att%apos - baselineAtt%apos) > maxDelta%apos) then 
          baselineAtt%apos = att%apos
          paIsNew = .true.
        endif

        if (raIsNew .or. decIsNew .or. paIsNew) then 
          baselineAtt = att
          attIsNew = .true.
        end if

        if (.not. lastAttWasGood) attIsNew = .true.
      end if

      noGoodAttsYet = .false.
      lastAttWasGood = .true.

    else ! att is not good
      if (lastAttWasGood .or. isFirstCall) then
        attIsNew = .true.
      else ! last att was no good
      end if

      lastAttWasGood = .false.
    end if

    isFirstCall = .false.

  end subroutine checkIfNewAttBin

  !---------------------------------------------------------------------
  subroutine calcMeanMedian(angles, isGood, meanVal, medianVal)
  !---------------------------------------------------------------------

    implicit none 

    real(double), intent(in)  :: angles(:)
    logical,      intent(in)  :: isGood(:)
    real(double), intent(out) :: meanVal, medianVal

    integer :: numGoodElements

    !-------------------------------------------------------------------
    ! Start:

    numGoodElements = count(isGood)
if (numGoodElements <= 0) call error('noElementsForMeanMedian',"")

    medianVal = median(angles, isGood) ! in ssclib/array_utils.
    meanVal = sum(angles, isGood) / real(numGoodElements, double)

  end subroutine calcMeanMedian

  !---------------------------------------------------------------------
  subroutine getEventAttitude(time, scAttitude, attIsGood)
  !---------------------------------------------------------------------
  ! NOTE! The OAL should be set up for attitudeFromAhf before this sub is called.

    implicit none 

    real(double),                 intent(in)  :: time
    type(SpacecraftAttitudeType), intent(out) :: scAttitude
    logical,                      intent(out) :: attIsGood

    !-------------------------------------------------------------------
    ! Start:

    if (OAL_getAttitude(time, scAttitude)) then 
      attIsGood = .true.
    else
      attIsGood = .false.
    end if

  end subroutine getEventAttitude

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module attbin_mod
